// Copyright (c) 2016, Christian Stewart. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'package:test/test.dart';

import 'package:julian/julian.dart';

Duration durationBetweenDates(DateTime from, DateTime to) => new Duration(
    milliseconds: to.millisecondsSinceEpoch - from.millisecondsSinceEpoch);

void main() {
  const int NOON = 12;
  final DateTime begin = new DateTime.utc(3000, DateTime.FEBRUARY, 1, NOON);
  final DateTime end = new DateTime.utc(-5000, DateTime.FEBRUARY, 1, NOON);
  // Julian day number is zero for November 24, 4714 BC (Gregorian)
  final DateTime epochGregorian =
      new DateTime.utc(-4713, DateTime.NOVEMBER, 24, NOON);
  DateTime current;
  int d;

  group('Gregorian, julianDayNumber', () {
    test('produces the correct Julian day number', () {
      expect(new Gregorian(0, 1, 1).julianDayNumber, equals(1721060));
    });
    test('produces the correct Jdn for a wide range of dates', () {
      d = 0;
      while ((current = begin.subtract(new Duration(days: d))).isAfter(end)) {
        expect(
            new Gregorian(current.year, current.month, current.day)
                .julianDayNumber,
            equals(durationBetweenDates(epochGregorian, current).inDays));
        d += 1;
      }
    });
  });

  group('Julian, julianDayNumber', () {
    test('produces the correct Julian day number', () {
      expect(new Julian(-4712, 1, 1).julianDayNumber, equals(0));
      expect(new Julian(-4712, 1, 2).julianDayNumber, equals(1));
      expect(new Julian(-4713, 12, 31).julianDayNumber, equals(-1));
    });
    test('separation between calendar dates is 13 days for year 2015', () {
      expect(
          new Julian(2015, 1, 1).julianDayNumber -
              new Gregorian(2015, 1, 1).julianDayNumber,
          equals(13));
      expect(new Julian(2015, 1, 1).julianDayNumber,
          equals(new Gregorian(2015, 1, 14).julianDayNumber));
    });
    test('produces the correct Jdn for a wide range of dates', () {
      d = 0;
      while ((current = begin.subtract(new Duration(days: d))).isAfter(end)) {
        Gregorian currentGregorian =
            new Gregorian(current.year, current.month, current.day);
        Julian currentJulian =
            new Julian.fromJulianDayNumber(currentGregorian.julianDayNumber);
        expect(
            new Gregorian.fromJulianDayNumber(currentJulian.julianDayNumber)
                .julianDayNumber,
            equals(durationBetweenDates(epochGregorian, current).inDays));
        d += 1;
      }
    });
  });

  group('switching between calendars', () {
    test('Julian 1582, October 5th and Gregorian 1582, October 15th', () {
      final Gregorian gregorianCalendarSwitchDay = new Gregorian(1582, 10, 15);
      final List<int> gregorianList = [1582, 10, 15];
      final Julian julianCalendarSwitchDay = new Julian(1582, 10, 5);
      final List<int> julianList = [1582, 10, 5];
      final DateTime switchDayAsDateTime = new DateTime(1582, 10, 15, 12);
      expect(
          julianCalendarSwitchDay
              .toGregorian()
              .toDateTime()
              .isAtSameMomentAs(switchDayAsDateTime),
          equals(true));
      expect(julianCalendarSwitchDay.toGregorian().toList(),
          equals(gregorianList));
      expect(
          gregorianCalendarSwitchDay.toJulian().toList(), equals(julianList));
      expect(new Julian.fromGregorian(gregorianCalendarSwitchDay).toList(),
          equals(julianList));
      expect(new Gregorian.fromJulian(julianCalendarSwitchDay).toList(),
          equals(gregorianList));
    });
  });
}
